import React, {useMemo, useRef, useState} from "react";
import {usePopper} from "react-popper";
import * as PopperJS from "@popperjs/core";
import {Card} from "../Card";

/**
 * Fully managed popover content function.
 */
export type Managed<T = React.ReactElement|React.ReactNode> = (show: () => void, hide: () => void) => T;

/**
 * Allowed popover modes.
 */
export type PopoverMode = "always"|"click"|"hover"|"focus"|"managed";

/**
 * A component to show something next to an element.
 */
export function Popover({children, content, className, mode, popperOptions}: {
	children: React.ReactElement|Managed<React.ReactElement>;
	content?: React.ReactNode|Managed<React.ReactNode>;
	className?: string;
	mode?: PopoverMode;
	popperOptions?: Partial<PopperJS.Options>;
}): React.ReactElement
{
	// By default, use "always" mode.
	if (!mode) mode = "always";

	// Followed show status.
	const [shown, setShown] = useState(false);

	// Create show / hide functions.
	const show = useMemo(() => (!shown ? () => setShown(true) : () => {}), [shown]);
	const hide = useMemo(() => (shown ? () => setShown(false) : () => {}), [shown]);

	// If show mode is "always", always show the popover after render.
	setTimeout(() => {
		if (mode == "always") setShown(true);
	}, 0);

	// HTML elements references.
	const referenceElement = useRef();
	const popperElement = useRef();

	// Change the child element to use the reference.
	const referencedChild = useMemo(() => (
		React.cloneElement(
			// Render the children if a managed popover function is passed.
			typeof children == "function" ? children(show, hide) : children,
			Object.assign(
				{
					ref: referenceElement,
				},
				// Pass click event if click mode is enabled.
				mode == "click" ? {
					onClick: (event: React.MouseEvent): void => {
						setShown(!shown);
					},
				} : {},
				// Pass focus events if focus mode is enabled.
				mode == "focus" ? {
					onFocus: show,
					onBlur: hide,
				} : {},
				// Pass the hover event if hover mode is enabled.
				mode == "hover" ? {
					onMouseEnter: show,
					onMouseOut: hide,
				} : {},
			))
	), [mode, shown, show, hide, children]);

	// Update popover content.
	const popoverContent = useMemo(() => (
		// Render the children if a managed popover function is passed.
		typeof content == "function" ? content(show, hide) : content
	), [shown, show, hide, content]);

	// Initialize popper.
	const { styles, attributes } = usePopper(
		referenceElement?.current, popperElement?.current, popperOptions
	);

	return (
		<>
			{referencedChild}

			<div ref={popperElement} style={styles.popper} {...attributes.popper} className={"popover"} hidden={!shown}>
				<Card className={`popover${className ? ` ${className}` : ""}`}>
					{popoverContent}
				</Card>
			</div>
		</>
	);
}
